# type: ignore
import os.path
import subprocess

import pytest
from rattler import Channel, ChannelConfig, fetch_repo_data, SparseRepoData, PackageName
from rattler.platform import Platform
from rattler.repo_data.record import RepoDataRecord


@pytest.fixture(scope="session")
def serve_repo_data() -> None:
    port, repo_name = 8912, "test-repo"

    test_data_dir = os.path.join(
        os.path.dirname(__file__), "../../../test-data/test-server"
    )

    with subprocess.Popen(
        [
            "python",
            os.path.join(test_data_dir, "reposerver.py"),
            "-d",
            os.path.join(test_data_dir, "repo"),
            "-n",
            repo_name,
            "-p",
            str(port),
        ]
    ) as proc:
        yield port, repo_name
        proc.terminate()


@pytest.mark.asyncio
async def test_fetch_repo_data(
    tmp_path,
    serve_repo_data,
):
    port, repo = serve_repo_data
    cache_dir = tmp_path / "test_repo_data_download"
    chan = Channel(repo, ChannelConfig(f"http://localhost:{port}/"))
    plat = Platform("noarch")

    result = await fetch_repo_data(
        channels=[chan],
        platforms=[plat],
        cache_path=cache_dir,
        callback=None,
    )
    assert isinstance(result, list)

    repodata = result[0]
    assert isinstance(repodata, SparseRepoData)

    package = PackageName(repodata.package_names()[0])
    repodata_record = repodata.load_records(package)[0]
    assert isinstance(repodata_record, RepoDataRecord)

    assert repodata_record.channel == f"http://localhost:{port}/{repo}/"
    assert repodata_record.file_name == "test-package-0.1-0.tar.bz2"
    assert str(repodata_record.package_record) == "test-package=0.1=0"
    assert (
        repodata_record.url
        == f"http://localhost:{port}/test-repo/noarch/test-package-0.1-0.tar.bz2"
    )



from __future__ import annotations
from typing import TYPE_CHECKING

from rattler.rattler import PyNamelessMatchSpec

if TYPE_CHECKING:
    from rattler.match_spec import MatchSpec
    from rattler.repo_data import PackageRecord


class NamelessMatchSpec:
    """
    Similar to a `MatchSpec` but does not include the package name.
    This is useful in places where the package name is already known
    (e.g. `foo = "3.4.1 *cuda"`).
    """

    def __init__(self, spec: str) -> None:
        if isinstance(spec, str):
            self._nameless_match_spec = PyNamelessMatchSpec(spec)
        else:
            raise TypeError(
                "NamelessMatchSpec constructor received unsupported type"
                f" {type(spec).__name__!r} for the 'spec' parameter"
            )

    def matches(self, package_record: PackageRecord) -> bool:
        """
        Match a MatchSpec against a PackageRecord
        """
        return self._nameless_match_spec.matches(package_record._package_record)

    @classmethod
    def _from_py_nameless_match_spec(
        cls, py_nameless_match_spec: PyNamelessMatchSpec
    ) -> NamelessMatchSpec:
        """
        Construct py-rattler NamelessMatchSpec from PyNamelessMatchSpec FFI object.
        """
        nameless_match_spec = cls.__new__(cls)
        nameless_match_spec._nameless_match_spec = py_nameless_match_spec

        return nameless_match_spec

    @classmethod
    def from_match_spec(cls, spec: MatchSpec) -> NamelessMatchSpec:
        """
        Constructs a NamelessMatchSpec from a MatchSpec.
        """
        return cls._from_py_nameless_match_spec(
            PyNamelessMatchSpec.from_match_spec(spec._match_spec)
        )

    def __str__(self) -> str:
        """
        Returns a string representation of the NamelessMatchSpec.
        """
        return self._nameless_match_spec.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the NamelessMatchSpec.
        """
        return f'NamelessMatchSpec("{self._nameless_match_spec.as_str()}")'



from __future__ import annotations
from rattler.rattler import PyAuthenticatedClient


class AuthenticatedClient:
    """
    A client that can be used to make authenticated requests.
    """

    def __init__(self) -> None:
        self._client = PyAuthenticatedClient()

    @classmethod
    def _from_ffi_object(cls, client: PyAuthenticatedClient) -> AuthenticatedClient:
        """
        Construct py-rattler AuthenticatedClient from PyAutheticatedClient FFI object.
        """
        authenticated_client = cls.__new__(cls)
        authenticated_client._client = client
        return authenticated_client

    def __repr__(self) -> str:
        """
        Returns a representation of the AuthenticatedClient

        Examples
        --------
        >>> AuthenticatedClient()
        AuthenticatedClient()
        """
        return f"{type(self).__name__}()"



from __future__ import annotations

from rattler.rattler import PyPackageRecord


class PackageRecord:
    """
    A single record in the Conda repodata. A single
    record refers to a single binary distribution
    of a package on a Conda channel.
    """

    def __init__(self) -> None:
        self._package_record = PyPackageRecord()

    @classmethod
    def _from_py_package_record(
        cls, py_package_record: PyPackageRecord
    ) -> PackageRecord:
        """
        Construct Rattler PackageRecord from FFI PyPackageRecord object.
        """
        package_record = cls.__new__(cls)
        package_record._package_record = py_package_record
        return package_record

    def __str__(self) -> str:
        """
        Returns the string representation of the PackageRecord.
        """
        return self._package_record.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the PackageRecord.
        """
        return f'PackageRecord("{self.__str__()}")'



from __future__ import annotations

from rattler.rattler import PyPackageName


class PackageName:
    def __init__(self, source: str) -> None:
        if not isinstance(source, str):
            raise TypeError(
                "PackageName constructor received unsupported type "
                f" {type(source).__name__!r} for the `source` parameter"
            )
        self._name = PyPackageName(source)

    @staticmethod
    def unchecked(normalized: str) -> PackageName:
        """
        Constructs a new `PackageName` from a string without checking if the string is actually a
        valid or normalized conda package name. This should only be used if you are sure that the
        input string is valid.

        Examples
        --------
        >>> p = PackageName.unchecked("test_xyz")
        """
        return PackageName._from_py_package_name(
            PyPackageName.new_unchecked(normalized)
        )

    @classmethod
    def _from_py_package_name(cls, py_package_name: PyPackageName) -> PackageName:
        """Construct Rattler PackageName from FFI PyPackageName object."""
        package_name = cls.__new__(cls)
        package_name._name = py_package_name
        return package_name

    @property
    def source(self) -> str:
        """
        Returns the source representation of the package name.
        This is the string from which this instance was created.

        Examples
        --------
        >>> p = PackageName("test-xyz")
        >>> p.source
        'test-xyz'
        """
        return self._name.source

    @property
    def normalized(self) -> str:
        """
        Returns the normalized version of the package name.
        The normalized string is guaranteed to be a valid conda package name.

        Examples
        --------
        >>> p = PackageName("test-xyz")
        >>> p.normalized
        'test-xyz'
        """
        return self._name.normalized

    def __repr__(self) -> str:
        """
        Returns a representation of the version.

        Examples
        --------
        >>> p = PackageName("test-xyz")
        >>> p
        PackageName("test-xyz")
        """
        return f'PackageName("{self.source}")'



from __future__ import annotations
import os
from pathlib import Path
from typing import List
from rattler.channel.channel import Channel
from rattler.package.package_name import PackageName

from rattler.rattler import PySparseRepoData
from rattler.repo_data.record import RepoDataRecord


class SparseRepoData:
    """
    A class to enable loading records from a `repodata.json` file on demand.
    Since most of the time you don't need all the records from the `repodata.json`
    this can help provide some significant speedups.
    """

    def __init__(
        self,
        channel: Channel,
        subdir: str,
        path: os.PathLike[str],
    ) -> None:
        if not isinstance(channel, Channel):
            raise TypeError(
                "SparseRepoData constructor received unsupported type "
                f" {type(channel).__name__!r} for the `channel` parameter"
            )
        if not isinstance(subdir, str):
            raise TypeError(
                "SparseRepoData constructor received unsupported type "
                f" {type(subdir).__name__!r} for the `subdir` parameter"
            )
        if not isinstance(path, (str, Path)):
            raise TypeError(
                "SparseRepoData constructor received unsupported type "
                f" {type(path).__name__!r} for the `path` parameter"
            )
        self._sparse = PySparseRepoData(channel._channel, subdir, path)

    def package_names(self) -> List[str]:
        """
        Returns a list over all package names in this repodata file.
        This works by iterating over all elements in the `packages` and
        `conda_packages` fields of the repodata and returning the unique
        package names.

        Examples
        --------
        >>> from rattler import Channel, ChannelConfig
        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> subdir = "test-data/conda-forge/noarch"
        >>> path = "../test-data/channels/conda-forge/noarch/repodata.json"
        >>> sparse_data = SparseRepoData(channel, subdir, path)
        >>> package_names = sparse_data.package_names()
        >>> package_names
        [...]
        >>> isinstance(package_names[0], str)
        True
        """
        return self._sparse.package_names()

    def load_records(self, package_name: PackageName) -> List[RepoDataRecord]:
        """
        Returns all the records for the specified package name.

        Examples
        --------
        >>> from rattler import Channel, ChannelConfig, RepoDataRecord, PackageName
        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> subdir = "test-data/conda-forge/noarch"
        >>> path = "../test-data/channels/conda-forge/noarch/repodata.json"
        >>> sparse_data = SparseRepoData(channel, subdir, path)
        >>> package_name = PackageName(sparse_data.package_names()[0])
        >>> records = sparse_data.load_records(package_name)
        >>> records
        [...]
        >>> isinstance(records[0], RepoDataRecord)
        True
        """
        # maybe change package_name to Union[str, PackageName]
        return [
            RepoDataRecord._from_py_record(record)
            for record in self._sparse.load_records(package_name._name)
        ]

    @property
    def subdir(self) -> str:
        """
        Returns the subdirectory from which this repodata was loaded.

        Examples
        --------
        >>> from rattler import Channel, ChannelConfig
        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> subdir = "test-data/conda-forge/noarch"
        >>> path = "../test-data/channels/conda-forge/noarch/repodata.json"
        >>> sparse_data = SparseRepoData(channel, subdir, path)
        >>> sparse_data.subdir
        'test-data/conda-forge/noarch'
        """
        return self._sparse.subdir

    @staticmethod
    def load_records_recursive(
        repo_data: List[SparseRepoData],
        package_names: List[PackageName],
    ) -> List[List[RepoDataRecord]]:
        """
        Given a set of [`SparseRepoData`]s load all the records
        for the packages with the specified names and all the packages
        these records depend on. This will parse the records for the
        specified packages as well as all the packages these records
        depend on.
        """
        return [
            [RepoDataRecord._from_py_record(record) for record in list_of_records]
            for list_of_records in PySparseRepoData.load_records_recursive(
                repo_data, package_names
            )
        ]

    @classmethod
    def _from_py_sparse_repo_data(
        cls, py_sparse_repo_data: PySparseRepoData
    ) -> SparseRepoData:
        """
        Construct Rattler SparseRepoData from FFI PySparseRepoData object.
        """
        sparse_repo_data = cls.__new__(cls)
        sparse_repo_data._sparse = py_sparse_repo_data
        return sparse_repo_data

    def __repr__(self) -> str:
        """
        Returns a representation of the SparseRepoData.

        Examples
        --------
        >>> from rattler import Channel, ChannelConfig
        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> subdir = "test-data/conda-forge/noarch"
        >>> path = "../test-data/channels/conda-forge/noarch/repodata.json"
        >>> sparse_data = SparseRepoData(channel, subdir, path)
        >>> sparse_data
        SparseRepoData(subdir="test-data/conda-forge/noarch")
        """
        return f'SparseRepoData(subdir="{self.subdir}")'



from __future__ import annotations

from typing import List, Optional, Tuple, Union

from rattler.rattler import PyVersion, InvalidVersionError


class Version:
    """
    This class implements an order relation between version strings.
    Version strings can contain the usual alphanumeric characters
    (A-Za-z0-9), separated into segments by dots and underscores.
    Empty segments (i.e. two consecutive dots, a leading/trailing
    underscore) are not permitted. An optional epoch number - an
    integer followed by '!' - can precede the actual version string
    (this is useful to indicate a change in the versioning scheme itself).
    Version comparison is case-insensitive.
    """

    def __init__(self, version: str) -> None:
        if isinstance(version, str):
            self._version = PyVersion(version)
        else:
            raise TypeError(
                "Version constructor received unsupported type "
                f" {type(version).__name__!r} for the `version` parameter"
            )

    @classmethod
    def _from_py_version(cls, py_version: PyVersion) -> Version:
        """Construct Rattler version from FFI PyVersion object."""
        version = cls.__new__(cls)
        version._version = py_version
        return version

    @property
    def epoch(self) -> Optional[str]:
        """
        Gets the epoch of the version or `None` if the epoch was not defined.

        Examples
        --------
        >>> v = Version('2!1.0')
        >>> v.epoch
        2
        """
        return self._version.epoch()

    def bump(self) -> Version:
        """
        Returns a new version where the last numerical segment of this version has
        been bumped.

        Examples
        --------
        >>> v = Version('1.0')
        >>> v.bump()
        Version("1.1")
        """
        return Version._from_py_version(self._version.bump())

    @property
    def has_local(self) -> bool:
        """
        Returns true if this version has a local segment defined.
        The local part of a version is the part behind the (optional) `+`.

        Examples
        --------
        >>> v = Version('1.0+3.2-alpha0')
        >>> v.has_local
        True
        >>> v2 = Version('1.0')
        >>> v2.has_local
        False
        """
        return self._version.has_local()

    def segments(self) -> List[List[Union[str, int]]]:
        """
        Returns a list of segments of the version. It does not contain
        the local segment of the version.

        Examples
        --------
        >>> v = Version("1.2dev.3-alpha4.5+6.8")
        >>> v.segments()
        [[1], [2, 'dev'], [3], [0, 'alpha', 4], [5]]
        """
        return self._version.segments()

    def local_segments(self) -> List[List[Union[str, int]]]:
        """
        Returns a list of local segments of the version. It does not
        contain the non-local segment of the version.

        Examples
        --------
        >>> v = Version("1.2dev.3-alpha4.5+6.8")
        >>> v.local_segments()
        [[6], [8]]
        """
        return self._version.local_segments()

    def as_major_minor(self) -> Optional[Tuple[int, int]]:
        """
        Returns the major and minor segments from the version.
        Requires a minimum of 2 segments in version to be split
        into major and minor, returns `None` otherwise.

        Examples
        --------
        >>> v = Version('1.0')
        >>> v.as_major_minor()
        (1, 0)
        """
        return self._version.as_major_minor()

    @property
    def is_dev(self) -> bool:
        """
        Returns true if the version contains a component name "dev",
        dev versions are sorted before non-dev version.

        Examples
        --------
        >>> v = Version('1.0.1dev')
        >>> v.is_dev
        True
        >>> v_non_dev = Version('1.0.1')
        >>> v_non_dev >= v
        True
        """
        return self._version.is_dev()

    def starts_with(self, other: Version) -> bool:
        """
        Checks if the version and local segment start
        same as other version.

        Examples
        --------
        >>> v1 = Version('1.0.1')
        >>> v2 = Version('1.0')
        >>> v1.starts_with(v2)
        True
        """
        return self._version.starts_with(other._version)

    def compatible_with(self, other: Version) -> bool:
        """
        Checks if this version is compatible with other version.
        Minor versions changes are compatible with older versions,
        major version changes are breaking and will not be compatible.

        Examples
        --------
        >>> v1 = Version('1.0')
        >>> v2 = Version('1.2')
        >>> v_major = Version('2.0')
        >>> v1.compatible_with(v2)
        False
        >>> v2.compatible_with(v1)
        True
        >>> v_major.compatible_with(v2)
        False
        >>> v2.compatible_with(v_major)
        False
        """
        return self._version.compatible_with(other._version)

    def pop_segments(self, n: int = 1) -> Version:
        """
        Pops `n` number of segments from the version and returns
        the new version. Raises `InvalidVersionError` if version
        becomes invalid due to the operation.

        Examples
        --------
        >>> v = Version('2!1.0.1')
        >>> v.pop_segments() # `n` defaults to 1 if left empty
        Version("2!1.0")
        >>> v.pop_segments(2) # old version is still usable
        Version("2!1")
        >>> v.pop_segments(3) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        exceptions.InvalidVersionException: new Version must have atleast 1 valid
        segment
        """
        new_py_version = self._version.pop_segments(n)
        if new_py_version:
            return self._from_py_version(new_py_version)
        else:
            raise InvalidVersionError("new Version must have atleast 1 valid segment")

    def with_segments(self, start: int, stop: int) -> Version:
        """
        Returns new version with with segments ranging from `start` to `stop`.
        `stop` is exclusive. Raises `InvalidVersionError` if the provided range
        is invalid.

        Examples
        --------
        >>> v = Version('2!1.2.3')
        >>> v.with_segments(0, 2)
        Version("2!1.2")
        """
        new_py_version = self._version.with_segments(start, stop)
        if new_py_version:
            return self._from_py_version(new_py_version)
        else:
            raise InvalidVersionError("Invalid segment range provided")

    @property
    def segment_count(self) -> int:
        """
        Returns the number of segments in the version.
        This does not include epoch or local segment of the version

        Examples
        --------
        >>> v = Version('2!1.2.3')
        >>> v.segment_count
        3
        """
        return self._version.segment_count()

    def strip_local(self) -> Version:
        """
        Returns a new version with local segment stripped.

        Examples
        --------
        >>> v = Version('1.2.3+4.alpha-5')
        >>> v.strip_local()
        Version("1.2.3")
        """
        return self._from_py_version(self._version.strip_local())

    def __str__(self) -> str:
        """
        Returns the string representation of the version

        Examples
        --------
        >>> str(Version("1.2.3"))
        '1.2.3'
        """
        return self._version.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the version

        Examples
        --------
        >>> Version("1.2.3")
        Version("1.2.3")
        """
        return f'Version("{self._version.as_str()}")'

    def __hash__(self) -> int:
        """
        Computes the hash of this instance.

        Examples
        --------
        >>> hash(Version("1.2.3")) == hash(Version("1.2.3"))
        True
        >>> hash(Version("1.2.3")) == hash(Version("3.2.1"))
        False
        >>> hash(Version("1")) == hash(Version("1.0.0"))
        True
        """
        return self._version.__hash__()

    def __eq__(self, other: Version) -> bool:  # type: ignore[override]
        """
        Returns True if this instance represents the same version as `other`.

        Examples
        --------
        >>> Version("1.2.3") == Version("1.2.3")
        True
        >>> Version("3.2.1") == Version("1.2.3")
        False
        >>> Version("1") == Version("1.0.0")
        True
        """
        return self._version == other._version

    def __ne__(self, other: Version) -> bool:  # type: ignore[override]
        """
        Returns True if this instance represents the same version as `other`.

        Examples
        --------
        >>> Version("1.2.3") != Version("1.2.3")
        False
        >>> Version("3.2.1") != Version("1.2.3")
        True
        >>> Version("1") != Version("1.0.0")
        False
        """
        return self._version != other._version

    def __gt__(self, other: Version) -> bool:
        """
        Returns True if this instance should be ordered *after* `other`.

        Examples
        --------
        >>> Version("1.2.3") > Version("1.2.3")
        False
        >>> Version("1.2.4") > Version("1.2.3")
        True
        >>> Version("1.2.3.1") > Version("1.2.3")
        True
        >>> Version("3.2.1") > Version("1.2.3")
        True
        >>> Version("1") > Version("1.0.0")
        False
        """
        return self._version > other._version

    def __lt__(self, other: Version) -> bool:
        """
        Returns True if this instance should be ordered *before* `other`.

        Examples
        --------
        >>> Version("1.2.3") < Version("1.2.3")
        False
        >>> Version("1.2.3") < Version("1.2.4")
        True
        >>> Version("1.2.3") < Version("1.2.3.1")
        True
        >>> Version("3.2.1") < Version("1.2.3")
        False
        >>> Version("1") < Version("1.0.0")
        False
        """
        return self._version < other._version

    def __ge__(self, other: Version) -> bool:
        """
        Returns True if this instance should be ordered *after* or at the same location
        as `other`.

        Examples
        --------
        >>> Version("1.2.3") >= Version("1.2.3")
        True
        >>> Version("1.2.4") >= Version("1.2.3")
        True
        >>> Version("1.2.3.1") >= Version("1.2.3")
        True
        >>> Version("3.2.1") >= Version("1.2.3")
        True
        >>> Version("1.2.3") >= Version("3.2.1")
        False
        >>> Version("1") >= Version("1.0.0")
        True
        """
        return self._version >= other._version

    def __le__(self, other: Version) -> bool:
        """
        Returns True if this instance should be ordered *before* or at the same
        location as `other`.

        Examples
        --------
        >>> Version("1.2.3") <= Version("1.2.3")
        True
        >>> Version("1.2.3") <= Version("1.2.4")
        True
        >>> Version("1.2.3") <= Version("1.2.3.1")
        True
        >>> Version("3.2.1") <= Version("1.2.3")
        False
        >>> Version("1") <= Version("1.0.0")
        True
        """
        return self._version <= other._version



from __future__ import annotations
from typing import Callable, List, Optional, Union, TYPE_CHECKING


from rattler.rattler import py_fetch_repo_data
from rattler.repo_data.sparse import SparseRepoData

if TYPE_CHECKING:
    import os
    from rattler.channel import Channel
    from rattler.platform import Platform


async def fetch_repo_data(
    *,
    channels: List[Channel],
    platforms: List[Platform],
    cache_path: Union[str, os.PathLike[str]],
    callback: Optional[Callable[[int, int], None]],
) -> List[SparseRepoData]:
    """
    Returns a list of RepoData for given channels and platform.
    """
    repo_data_list = await py_fetch_repo_data(
        [channel._channel for channel in channels],
        [platform._inner for platform in platforms],
        cache_path,
        callback,
    )

    return [
        SparseRepoData._from_py_sparse_repo_data(repo_data)
        for repo_data in repo_data_list
    ]



from __future__ import annotations

from rattler.rattler import PyRepoDataRecord
from rattler.repo_data import PackageRecord


class RepoDataRecord:
    _record: PyRepoDataRecord

    @property
    def package_record(self) -> PackageRecord:
        """
        The data stored in the repodata.json.
        """
        return PackageRecord._from_py_package_record(self._record.package_record)

    @property
    def url(self) -> str:
        """
        The canonical URL from where to get this package.
        """
        return self._record.url

    @property
    def channel(self) -> str:
        """
        String representation of the channel where the
        package comes from. This could be a URL but it
        could also be a channel name.
        """
        return self._record.channel

    @property
    def file_name(self) -> str:
        """
        The filename of the package.
        """
        return self._record.file_name

    @classmethod
    def _from_py_record(cls, py_record: PyRepoDataRecord) -> RepoDataRecord:
        """
        Construct Rattler RepoDataRecord from FFI PyRepoDataRecord object.
        """
        record = cls.__new__(cls)
        record._record = py_record
        return record

    def __repr__(self) -> str:
        """
        Returns a representation of the RepoDataRecord.
        """
        return f"{type(self).__name__}()"



from __future__ import annotations
from typing import Optional

from rattler.rattler import PyChannel
from rattler.channel.channel_config import ChannelConfig


class Channel:
    def __init__(
        self, name: str, channel_configuration: ChannelConfig = ChannelConfig()
    ) -> None:
        """
        Create a new channel.

        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> channel
        Channel(name="conda-forge", base_url="https://conda.anaconda.org/conda-forge/")
        """
        self._channel = PyChannel(name, channel_configuration._channel_configuration)

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of this channel.

        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> channel.name
        'conda-forge'
        """
        return self._channel.name

    @property
    def base_url(self) -> str:
        """
        Return the base URL of this channel.

        >>> channel = Channel("conda-forge", ChannelConfig())
        >>> channel.base_url
        'https://conda.anaconda.org/conda-forge/'
        """
        return self._channel.base_url

    def __repr__(self) -> str:
        """
        Return a string representation of this channel.
        """
        return f'Channel(name="{self.name}", base_url="{self.base_url}")'



import pytest
from rattler import Version


def test_version_dash_normalisation() -> None:
    assert Version("1.0-").segments() == [[1], [0, "_"]]
    assert Version("1.0_").segments() == [[1], [0, "_"]]
    assert Version("1.0dev-+2.3").segments() == [[1], [0, "dev", "_"]]
    assert Version("1.0dev_").segments() == [[1], [0, "dev", "_"]]

    assert Version("1.0dev-+2.3").local_segments() == [[2], [3]]
    assert Version("1.0dev+3.4-dev").local_segments() == [[3], [4], [0, "dev"]]
    assert Version("1.0dev+3.4-").local_segments() == [[3], [4, "_"]]

    with pytest.raises(Exception):
        Version("1-.0dev-")

    with pytest.raises(Exception):
        Version("1-.0dev+3.4-")



from __future__ import annotations
from typing import Literal

from rattler.rattler import PyPlatform
from rattler.platform.arch import Arch

PlatformLiteral = Literal[
    "noarch",
    "linux-32",
    "linux-64",
    "linux-aarch64",
    "linux-armv6l",
    "linux-armv7l",
    "linux-ppc64le",
    "linux-ppc64",
    "linux-s390x",
    "linux-riscv32",
    "linux-riscv64",
    "osx-64",
    "osx-arm64",
    "win-32",
    "win-64",
    "win-arm64",
    "emscripten-32",
]


class Platform:
    def __init__(self, value: PlatformLiteral):
        self._inner = PyPlatform(value)

    @classmethod
    def _from_py_platform(cls, py_platform: PyPlatform) -> Platform:
        """Construct Rattler version from FFI PyArch object."""
        platform = cls.__new__(cls)
        platform._inner = py_platform
        return platform

    def __str__(self) -> str:
        """
        Returns a string representation of the platform.

        >>> str(Platform("linux-64"))
        'linux-64'
        """
        return self._inner.name

    def __repr__(self) -> str:
        """
        Returnrs a representation of the platform.

        >>> Platform("linux-64")
        Platform(linux-64)
        """
        return f"Platform({self._inner.name})"

    @classmethod
    def current(cls) -> Platform:
        """
        Returns the current platform.

        # >>> Platform.current()
        # Platform(linux-64)
        """
        return Platform._from_py_platform(PyPlatform.current())

    @property
    def is_linux(self) -> bool:
        """
        Return True if the platform is linux.

        >>> Platform("linux-64").is_linux
        True
        >>> Platform("osx-64").is_linux
        False
        """
        return self._inner.is_linux

    @property
    def is_osx(self) -> bool:
        """
        Return True if the platform is osx.

        >>> Platform("osx-64").is_osx
        True
        >>> Platform("linux-64").is_osx
        False
        """
        return self._inner.is_osx

    @property
    def is_windows(self) -> bool:
        """
        Return True if the platform is win.

        >>> Platform("win-64").is_windows
        True
        >>> Platform("linux-64").is_windows
        False
        """
        return self._inner.is_windows

    @property
    def is_unix(self) -> bool:
        """
        Return True if the platform is unix.

        >>> Platform("linux-64").is_unix
        True
        >>> Platform("win-64").is_unix
        False
        """
        return self._inner.is_unix

    @property
    def arch(self) -> Arch:
        """
        Return the architecture of the platform.

        >>> Platform("linux-64").arch
        Arch(x86_64)
        >>> Platform("linux-aarch64").arch
        Arch(aarch64)
        """
        return Arch._from_py_arch(self._inner.arch())



from __future__ import annotations
import os
from typing import List

from rattler.rattler import PyPrefixPaths


class PrefixPaths:
    _paths: PyPrefixPaths

    @classmethod
    def _from_py_prefix_paths(cls, py_prefix_paths: PyPrefixPaths) -> PrefixPaths:
        """Construct Rattler PrefixRecord from FFI PyPrefixRecord object."""
        paths = cls.__new__(cls)
        paths._paths = py_prefix_paths
        return paths

    @property
    def paths_version(self) -> int:
        """
        The version of the file.

        Examples
        --------
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> paths.paths_version
        1
        """
        return self._paths.paths_version

    @property
    def paths(self) -> List[os.PathLike[str]]:
        """
        All entries included in the package.

        Examples
        --------
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> paths = r.paths_data
        >>> paths.paths # doctest:+ELLIPSIS
        [...]
        """
        return self._paths.paths

    def __repr__(self) -> str:
        """
        Returns a representation of the PrefixPaths.

        Examples
        --------
        >>> from rattler.prefix.prefix_record import PrefixRecord
        >>> r = PrefixRecord.from_path(
        ...     "../test-data/conda-meta/requests-2.28.2-pyhd8ed1ab_0.json"
        ... )
        >>> r.paths_data
        PrefixPaths()
        """
        return "PrefixPaths()"



from __future__ import annotations

from rattler.rattler import PyChannelConfig


class ChannelConfig:
    def __init__(self, channel_alias: str = "https://conda.anaconda.org/") -> None:
        """
        Create a new channel configuration.

        >>> channel_config = ChannelConfig()
        >>> channel_config
        ChannelConfig(channel_alias="https://conda.anaconda.org/")

        >>> channel_config = ChannelConfig("https://repo.prefix.dev/")
        >>> channel_config
        ChannelConfig(channel_alias="https://repo.prefix.dev/")
        """
        self._channel_configuration = PyChannelConfig(channel_alias)

    def __repr__(self) -> str:
        """
        Return a string representation of this channel configuration.
        """
        alias = self._channel_configuration.channel_alias
        return f'ChannelConfig(channel_alias="{alias}")'



from __future__ import annotations

from rattler.rattler import PyVirtualPackage
from typing import List

from rattler.virtual_package.generic import GenericVirtualPackage


class VirtualPackage:
    _virtual_package: PyVirtualPackage

    @classmethod
    def _from_py_virtual_package(
        cls, py_virtual_package: PyVirtualPackage
    ) -> VirtualPackage:
        """Construct Rattler VirtualPackage from FFI PyVirtualPackage object."""
        virtual_package = cls.__new__(cls)
        virtual_package._virtual_package = py_virtual_package
        return virtual_package

    @staticmethod
    def current() -> List[VirtualPackage]:
        """
        Returns virtual packages detected for the current system or an error
        if the versions could not be properly detected.

        Examples
        --------
        >>> v = VirtualPackage.current()
        """
        return [
            VirtualPackage._from_py_virtual_package(vp)
            for vp in PyVirtualPackage.current()
        ]

    def into_generic(self) -> GenericVirtualPackage:
        """
        Returns a GenericVirtualPackage from VirtualPackage.
        """
        # subclass from Generic instead.
        return GenericVirtualPackage._from_py_generic_virtual_package(
            self._virtual_package.as_generic()
        )

    def __str__(self) -> str:
        """
        Returns string representation of the VirtualPackage.
        """
        return self._virtual_package.as_str()

    def __repr__(self) -> str:
        """
        Returns a representation of the VirtualPackage.
        """
        return f"VirtualPackage({self._virtual_package.as_str()})"



from __future__ import annotations
from pathlib import Path
from typing import Union, List, TYPE_CHECKING


if TYPE_CHECKING:
    from os import PathLike
    from rattler.channel import Channel
    from rattler.repo_data import PatchInstructions, RepoDataRecord

from rattler.rattler import PyRepoData


class RepoData:
    def __init__(self, path: Union[str, PathLike[str]]) -> None:
        if not isinstance(path, (str, Path)):
            raise TypeError(
                "RepoData constructor received unsupported type "
                f" {type(path).__name__!r} for the `path` parameter"
            )

        self._repo_data = PyRepoData.from_path(path)

    def apply_patches(self, instructions: PatchInstructions) -> None:
        """
        Apply a patch to a repodata file.
        Note that we currently do not handle revoked instructions.
        """
        self._repo_data.apply_patches(instructions._patch_instructions)

    def into_repo_data(self, channel: Channel) -> List[RepoDataRecord]:
        """
        Builds a `List[RepoDataRecord]` from the packages in a
        `RepoData` given the source of the data.
        """
        from rattler.repo_data import RepoDataRecord

        return [
            RepoDataRecord._from_py_record(record)
            for record in PyRepoData.repo_data_to_records(
                self._repo_data, channel._channel
            )
        ]

    @classmethod
    def _from_py_repo_data(cls, py_repo_data: PyRepoData) -> RepoData:
        """
        Construct Rattler RepoData from FFI PyRepoData object.
        """
        repo_data = cls.__new__(cls)
        repo_data._repo_data = py_repo_data
        return repo_data

    def __repr__(self) -> str:
        """
        Returns a representation of the RepoData.
        """
        return f"{type(self).__name__}()"
